/**************************************************************
*
*    FTT-CAN Version 9
*
*     Master station (arbitrator)
*
*  scans the arbitrator variables' table
*  invoke the planning scheduler every NEC
*    elementary cycles to build a plan
*  invoke the dispatcher every EC to send
*    master message indicating the variables
*    to be produced during that EC
*
*  History:
*    98/9/30 - created by Luis de Almeida
*              using CAN libs developed by Pedro Fonseca  
*    98/10/1 - use of unsigned chars maximized
*    98/10/13 - collect timing data concerning:
*                 - duration of dispatcher
*                 - duration of scheduler
*              use a buffer to store data, when full:
*                 - send all data via RS232
*                 - terminate 
*              uses the same timer as to generate EC ints
*                (resolution +/- 32 us)
*    98/10/15 - plan tables contain ECs already codified
*                as master message data (this part moved
*                from the dispatcher to the scheduler)
*    98/10/19 - ECs load goes from 0 to EC_MAX_bits
*    98/10/20 - prevIndex in EcType used to save the order
*                   in the variables' table of the last var added
*                   to each EC.
*    98/11/6 -  new model to calculate max production delay
*                   transmissions are started after a fixed SCAN_WINDOW
*    99/1/12 -  now uses header file plansch5.h
*    99/1/18 -  function to add vars to vars' table has been added here
*    99/2/17 -  version 6 created, essentially equal to version 5
*                   differences in the management of the messages' table
*   00/02/01 -  Version 7 developed jointly by Luis Almeida & Paulo Pedreiras.
*                   Scanning window size dynamicaly adjusted in each EC so the 
*                   synchronous message area varies accordingly to the number and
*	                size of the synchronous messages produced in each EC,being
*                   at the end of each EC.
*   00/10/23 - Version 8 developed jointly by L. Almeida and P. Pedreiras.
*              		i) Production of asynchronous messages under controlled or 
*						uncontrolled mode;
*   00/01/25 - Scheduler can be Rate Monotonic (as before) or EDF.
*              define directive EDF/RM to select.
*  02/01/29 - Version 9 developed by L. Almeida, P. Pedreiras and J. Ferreira
*							Error handling added:
*							- Master redundancy
*							- Master synchronisation
*							- Message error handling
*
***************************************************************************************/

/* Includes ****************************************/
#include "io592.h"
#include <80592.h>
#include "plansch9.h"
#include <special.h>
#include <math.h>

/* Defines *****************************************/
#define RM 				/* RM or EDF Scheduler 	*/
#define MSTVER 9		/* Software version 		*/
#define MST_ID 0x02		/* Master ID (0..7) 		*/

/* Function Prototypes *****************************/
void ini_globals(void);
void ini_CAN(void);

unsigned char p_scheduler(PlanType *);
signed char MstTab_AddVar(MstVarType *);

unsigned char msg_set_add(void); /* Adds a set of messages to the SRT */
unsigned char get_network_status(void); /* Returns the status of the network */
unsigned char send_mst_data();	/* Sends active master data messages */

void Send_DBG_Mesg();

/****************/
/* Global vars: */
/****************/

PlanType 	plan_1,
	  			plan_2;		/* Real plans */
PlanType 	*sched_plan,
	  			*disp_plan,
	  			*temp_plan;   /* Pointers to plans in use */

MstVTabType arb_table; /* Table with message properties */

data unsigned char CurrentEC;  /* EC being processed by the dispatcher */

data unsigned char mstdata[MAX_VAR/8 + 1];	/* Master message data - Note: first data byte of the trigger message conveys*/
																						/* 	the sync. window length in slots (500us) */
data unsigned char Dynamic_Sync_Window;	/* Signals if asynchronous messages are supported */

/* Timing data buffers and variables */
data unsigned char NintECs;
data unsigned int SchedTIndex;
data unsigned int DispTIndex;
#define DISPBUFSIZ 3000  /* o Macro-ciclo tem periodo de 2520 */
#define SCHEDBUFSIZ DISPBUFSIZ/NEC
unsigned char SchedTime[SCHEDBUFSIZ][3];
unsigned char DispTime[DISPBUFSIZ][2];
data unsigned char sch_TH0i,sch_TH0f;
data unsigned int Cur_NintECs;

/* General variables */
data unsigned char ec_count;
data unsigned int lb,ls500;
data unsigned char Master_Status; 	/* Master status : active or backup */
data unsigned char Pol_Status;		/* Backup Master Pol Status */
data unsigned char IntMesgBuf[10];  /* internal message buffer (full frame) for Rx */
data unsigned char FTT_EC_detected;	/* Flag to signal the detection of a valid EC in the network */
data unsigned char Master_SendSRT_Data; /* Flag to signal that a backup master has requested the SRT data */
data unsigned char Master_SendInd;	/* Counter used to control the queueing of SRT data */

data unsigned char NewPlan_Flag;		/* Signals that a new plan has begin */

data unsigned char tm_seqnum;				/* Sequence number of trigger message */

data unsigned char TH0_asylen;		/* Holds the value that Timer 0 counter will have at the end of the async. window */

data unsigned char ctrl_mesg_data[8]; /* Data carried in control messages */

data unsigned int LTM_min_us; 	/* Minimum tx time in microsecs of the TM (i.e., assuming no stuff bits) */
data unsigned int TM_ETE_delay_us, 	/* End-to-end delay for TM transmission in usecs and TH0 ticks */
									TM_ETE_delay_TH0_ticks;

MstDataQueueType MstDataQueue;	/* Queue to store received master data */

/* Variables used in S0_Int - CAN Int handler */
data unsigned char i,nbm;	/* general counter */
data unsigned int cm_type; /* Type of control message */

/* Global vars used by function send_mst_data */
data unsigned char cur_ind,
				   TotMesgs_MsgProp,
				   TotMesgs_SchInf,
					 Master_SendInd_TotMesgs; /* Used for control of the process of sending SRT data to a backup master */
data unsigned char SchedRun_Flag;	/* Signal if scheduler is running */

/* Debug variables */
data unsigned char dbg_t0_cnt;
unsigned char dbg_wait_cnt;
data unsigned char dbg_parar;
data unsigned int dbg_parar_cnt;

data unsigned char dbg_info;


/**************************************************
*
*   main
*
*  initiate timer and globals
*  build variables' table
*  setup cyclical activities
*  infinite loop including call to scheduler
*
*  construction of plans and data transmission
*  stop upon detection of any abnormal situation
*
*  abnormal situations:
*    deadline missed when building a plan
*    scheduler takes too long to build the plan
*    
*
****************************************************/
void main(void)
{
	unsigned char net_stat; /* Status of the newtork */
	unsigned char msg_ind;
	MstVarType aux_var; /* Auxiliary variable used to update SRT */
	unsigned char msgprop_ind, schinf_ind, msgprop_bind, schinf_bind; /* Indexes to extract master data from Queue */

								
	

	/* Inits - CAN, timer and global variables */
	ini_globals();

	#ifdef RM
		printf("\n Starting Master 9 in Rate Monotonic (RM) mode [%d]...\n",MSTVER);
	#endif

	#ifdef EDF
		printf("\n Starting Master 9 in Earliest Deadline First (EDF) mode [%d]...\n",MSTVER);
	#endif

	
	/* Set port 4 status (for debug) */
	P4.7 = 0; /* pulse for external sync */
	P4.5 = 0;

	while(TRUE)
	{
		/* Get network status: */
		/* 	If no network activity, configure a predefined set of messages and become the Active Master */
		/*	else, get scheduling info from the current active master and become a Backup Master */
  	net_stat=get_network_status();
		if(net_stat == NETWORK_IDLE)
		{	
			
			  /******************************************************************************/
			 /* Network is idle - load a (predefined) message set and become Active Master */
			/******************************************************************************/

			dbg_parar=FALSE;
			dbg_parar_cnt=0;

			Master_Status = ACTIVE_MASTER;
			timer_arranque(0,1);	/* Start Scheduler */
			msg_set_add(); /* Adds messages to the SRT */

			/* Initialize plans for cyclical activity */
			if( p_scheduler( sched_plan ) == MISSED_DEADLINE )
			{   /* deadline missed, disable ints and stop */

				P4.7 = 1;
				IEN0 = 0;
				while( 1 ) {}; /* ad eternum */
			}

			/* Sched_plan->status is now UPDATED */
			/* 		dispatcher is idle while disp_plan->status = OUT_OF_DATE */
			/* 		switch plans, causing the dispatcher to start */
			temp_plan = sched_plan;
			sched_plan = disp_plan;
			disp_plan = temp_plan;  /* set disp_plan last because of ints */

			   /****************************************************************/
			  /* Main cycle: 																									*/
			 /* 	wait while the dipatcher processes one plan 								*/
			/*  and, after that, execute the scheduler to build a new plan 	*/
			/****************************************************************/
			P4.7 = 0;
			P4.6 = 0;
			ec_count=0;
			while( TRUE ) /* Forever */
			{
				/* Wait for dispatcher to finish the plan and return it with status==OUT_OF_DATE */
				while( sched_plan->status != OUT_OF_DATE)
				{				
					if(Master_Status == BACKUP_MASTER_OUTSYNC)
						break; /* If Master had timing problems */

				} /* while(sched_plan->status != OUT_OF_DATE) */
			
				if(Master_Status == BACKUP_MASTER_OUTSYNC)
						break; /* If Master had timing problems */
		
				/* Build a new plan (the dispatcher switches between them) */
				if( p_scheduler( sched_plan ) == MISSED_DEADLINE )
				{
					/* If deadline missed, disable ints and stop */
					P4.7 = 1;
					IEN0 = 0;
					while( 1 ) {};                   /* ad eternum */
				}
			}
	
		}
		else
		{	
			  /*****************************************************************/
			 /* Normal FTT network activity detected - become a Backup Master */
			/*****************************************************************/
			
			/* Init some variables rquired by the synch process */
			LTM_min_us = MAX_VAR / 8;
			if(LTM_min_us * 8 < MAX_VAR) LTM_min_us++; /* Get number of data bytes required */
			LTM_min_us *= 8; /* Convert to data bits */
			LTM_min_us += (bits_OVRHD_sb + bits_OVRHD_nsb); /* Add CAN overhead bits SOF,RTR,DLC ... */
			LTM_min_us = ((unsigned long)(LTM_min_us)*1000/TX_RATE); /* Convert bits to us @ TX_RATE */

			TM_ETE_delay_us = TMDIS_OVR_us + LTM_min_us + TDEC_OVR_us;
			TM_ETE_delay_TH0_ticks = TM_ETE_delay_us / 34.75;

			printf("\n TM Delay (MAX_VAR=%d) = %dus",MAX_VAR,TM_ETE_delay_TH0_ticks);

			/* Init queue data structure: index of received Master data and others */
			MstDataQueue.nmesgs = 0;
			MstDataQueue.nm_srt = 0;
			MstDataQueue.nm_msgprop = 0;
			MstDataQueue.nm_schinf = 0;		
		
			Master_Status = BACKUP_MASTER_OUTSYNC; /* Set up flags to send Query message and receive master data */
			Pol_Status = BM_POL_DATA;
		
			/* Get message properties ... */
			while(Pol_Status != BM_ROK_DATA);

			/* DEBUG CODE */
			/*
			dbg_info=1;
			Send_DBG_Mesg();
			*/
			/***************/

			/* Update SRT with received data */
			for(msg_ind=0;msg_ind<MstDataQueue.nm_srt;msg_ind++)
			{
				/* Calculate Queue index and byte index inside message */
				msgprop_ind = msg_ind / 2;
				msgprop_bind = (msg_ind % 2) * 4;
						
				schinf_ind = msg_ind / 4 + MstDataQueue.nm_msgprop;
				schinf_bind = (msg_ind % 4) * 2;
		
				aux_var.vid=MstDataQueue.mesg_data[msgprop_ind][msgprop_bind+3];
				aux_var.size=MstDataQueue.mesg_data[msgprop_ind][msgprop_bind+2];
				aux_var.period=MstDataQueue.mesg_data[msgprop_ind][msgprop_bind+1];
				aux_var.deadline=MstDataQueue.mesg_data[msgprop_ind][msgprop_bind];
				aux_var.init=MstDataQueue.mesg_data[schinf_ind][schinf_bind];
				MstTab_AddVar(&aux_var);

			}

			/* DEBUG CODE */
			/*
			dbg_info=2;
			Send_DBG_Mesg();
			*/
			/***************/

			/****DEBUG******/
			/*
			EA=0;
			for(msg_ind=0;msg_ind<arb_table.nvars;msg_ind++)
			{
				printf("\n---------");
				printf("\n ID: %d",arb_table.vars[msg_ind].vid);
				printf("\n 		period	: %d",arb_table.vars[msg_ind].period);
				printf("\n 		deadline: %d",arb_table.vars[msg_ind].deadline);
				printf("\n 		size		: %d",arb_table.vars[msg_ind].size);
				printf("\n 		init		: %d",arb_table.vars[msg_ind].init);
			}

			while(1);
			*/
			/***************/


			/* Wait for new plan Note: Flag was set to false on reception of MST_DATA_OK */
			while(NewPlan_Flag == FALSE);

			/* DEBUG CODE */
			/*
			dbg_info=3;
			Send_DBG_Mesg();
			*/
			/***************/

			/* Set-up flag in order to detect next plan */
			NewPlan_Flag = FALSE;
		
			/* Build next plan  */
			if( p_scheduler( sched_plan ) == MISSED_DEADLINE )
			{   /* deadline missed, disable ints and stop */
				P4.7 = 1;
				IEN0 = 0;
				while( 1 ) {}; /* ad eternum */
			}
		
			/* Sched_plan->status is now UPDATED */
			temp_plan = sched_plan;
			sched_plan = disp_plan;
			disp_plan = temp_plan;
    	
			/* Wait for new plan*/
			while(NewPlan_Flag == FALSE);

			/* Synchronize with active master */
			Master_Status = BACKUP_MASTER_GETSYNC; /* Set up flag to get Timer and Dispatcher sync. with active Master */

			/* DEBUG CODE */
			/*
			dbg_info=4;
			Send_DBG_Mesg();
			*/
			/***************/

			/* Build next plan  */
			if( p_scheduler( sched_plan ) == MISSED_DEADLINE )
			{   /* deadline missed, disable ints and stop */
				P4.7 = 1;
				IEN0 = 0;
				while( 1 ) {}; /* ad eternum */
			}

			dbg_t0_cnt=0;

			while( TRUE ) /* Forever */
			{
				printf(".");
				/* Wait for dispatcher to finish the plan and return it with status==OUT_OF_DATE */
				while( sched_plan->status != OUT_OF_DATE)
				{	
					if(Master_Status == BACKUP_MASTER_OUTSYNC)
						break; /* If Master had timing problems */
				} /* while(sched_plan->status != OUT_OF_DATE) */

				if(Master_Status == BACKUP_MASTER_OUTSYNC)
						break; /* If Master had timing problems */
		
				/* Build a new plan (the dispatcher switches between them) */
				if( p_scheduler( sched_plan ) == MISSED_DEADLINE )
				{
					/* If deadline missed, disable ints and stop */
					P4.7 = 1;
					IEN0 = 0;
					while( 1 ) {};                   /* ad eternum */
				}
			}
		
		} /* IF NETWORK_STATUS ... */
		
	}/* WHILE(TRUE) */


}

/******************************************************************************
*																																							*
*    <<<<<<<<<<<Function implementation>>>>>>>>>>>>                           *
*                                                                             *
*******************************************************************************/

/***************************************************************************
*
* send_mst_data
*
* Sends a control message (in the asynchronous transmition window) with active master's data
* Inputs:
*       Global vars:
*
*   Outputs:
*       - None
*
*		Changes global vars:
*				
*
****************************************************************************/
unsigned char send_mst_data()
{
	unsigned char nb_ctrl_mesg,i;


	/*******************************************************/
	/* Check if beginning of queueing of SRT data messages */
	/*******************************************************/
	if(Master_SendInd == 0)
	{

		/* Calculate number of messages that will be queued */
		TotMesgs_MsgProp = arb_table.nvars / 2;	/* Number of MST_DATA_MSGPROP messages */
		if(arb_table.nvars != (TotMesgs_MsgProp * 2) )
			TotMesgs_MsgProp++;

		TotMesgs_SchInf = arb_table.nvars / 4;	/* Number of MST_DATA_SCHINF messages */
		if(arb_table.nvars != (TotMesgs_SchInf * 4) )
			TotMesgs_SchInf++;

		Master_SendInd_TotMesgs = TotMesgs_MsgProp + TotMesgs_SchInf;
		cur_ind = 0;	/* index for message in the SRT */

		/*Txmit message signaling the beginning of tx of Master Data*/
		CANADR = CAN_TX | AutoInc;

		/* 1st byte - ID10..ID3 */
		CANDAT = (CTRL_MESG_ID_HP | MST_DATA_MSGPROP_BEG) >> 3;
		
		/* 2nd byte : ID2..ID0 + RTR + DLC */
		CANDAT = ( MST_DATA_MSGPROP_BEG << 5 ) | ( 0x03);

		/* First data byte - Number of synchronous messages on the active master */
		CANDAT = arb_table.nvars;

		/* 2nd data byte - Number of mesgs MSGPROP */
		CANDAT = TotMesgs_MsgProp;

		/* 3rd data byte - Number of mesgs SCHINF */
		CANDAT = TotMesgs_SchInf;

		/* Set TR bit (Transmission Request) */
		CANADR = CAN_CMR;
		CANDAT &= 0xF1;    /* keep high 4 bits, TR is read as 1, reset others */

		Master_SendInd = 1;

		return 0;

	}

	/******************************************/

	/* Check if all messages were transmitted */
	/******************************************/
	if(Master_SendInd == (Master_SendInd_TotMesgs+1))
	{
		if(NewPlan_Flag == FALSE) /* Info sent is accurate */
		{
			Master_SendSRT_Data = FALSE;
			
			/*Txmit message signaling data txfer complete and OK*/
			CANADR = CAN_TX | AutoInc;

			/* 1st byte - ID10..ID3 */
			CANDAT = (CTRL_MESG_ID_HP | MST_DATA_OK) >> 3;
		
			/* 2nd byte : ID2..ID0 + RTR + DLC */
			CANDAT = ( MST_DATA_OK << 5 ) | ( 0x00);

			/* Set TR bit (Transmission Request) */
			CANADR = CAN_CMR;
			CANDAT &= 0xF1;    /* keep high 4 bits, TR is read as 1, reset others */

			dbg_parar = TRUE;
			dbg_parar_cnt=0;

			return 0;
		}
		else /* New plan during data txmit -> repeat SchInf mesgs with updated data */
		{
			Master_SendInd = TotMesgs_MsgProp+1; /* Reset index to schinf data */
			
			/*Txmit message signaling data txfer repetition (SCHINF only) */
			CANADR = CAN_TX | AutoInc;

			/* 1st byte - ID10..ID3 */
			CANDAT = (CTRL_MESG_ID_HP | MST_DATA_SCHINF_REFRESH) >> 3;
		
			/* 2nd byte : ID2..ID0 + RTR + DLC */
			CANDAT = ( MST_DATA_SCHINF_REFRESH << 5 ) | ( 0x00);

			/* Set TR bit (Transmission Request) */
			CANADR = CAN_CMR;
			CANDAT &= 0xF1;    /* keep high 4 bits, TR is read as 1, reset others */
			
			return 0;
	
		}
	}

	/*****************************************************/
	/* Check if next message is of type MST_DATA_MSGPROP */
	/*****************************************************/
	if(Master_SendInd <= TotMesgs_MsgProp)
	{
		/* Store message properties in buffer */
		ctrl_mesg_data[0] = arb_table.vars[cur_ind].deadline;
		ctrl_mesg_data[1] = arb_table.vars[cur_ind].period;
		ctrl_mesg_data[2] = arb_table.vars[cur_ind].size;
		ctrl_mesg_data[3] = arb_table.vars[cur_ind].vid;
		cur_ind++;
		nb_ctrl_mesg=4;
		
		if(cur_ind<arb_table.nvars)
		{	/* If there are more messages(there is place for 2 in each MST_DATA_MSGPROP frame */
			ctrl_mesg_data[4] = arb_table.vars[cur_ind].deadline;
			ctrl_mesg_data[5] = arb_table.vars[cur_ind].period;
			ctrl_mesg_data[6] = arb_table.vars[cur_ind].size;
			ctrl_mesg_data[7] = arb_table.vars[cur_ind].vid;
			cur_ind++;
			nb_ctrl_mesg=8;
		}
		
		/* Txmit control message	*/
		CANADR = CAN_TX | AutoInc;
		/* 1st byte - ID10..ID3 */	
		CANDAT = (CTRL_MESG_ID_HP | MST_DATA_MSGPROP) >> 3;
		
		/* 2nd byte : ID2..ID0 + RTR + DLC */
		CANDAT = ( MST_DATA_MSGPROP << 5 ) | nb_ctrl_mesg;
	
		/* Queue data bytes */
		for(i=0;i<nb_ctrl_mesg;i++)		
			CANDAT = ctrl_mesg_data[i];

		/* Set TR bit (Transmission Request) */
		CANADR = CAN_CMR;
		CANDAT &= 0xF1;    /* keep high 4 bits, TR is read as 1, reset others */

		Master_SendInd++; /* Update index */

		return 0;
	}
	else
	{
		/****************************************************/
		/* Check if next message is of type MST_DATA_SCHINF */
		/****************************************************/

		/* If scheduler is still running, can't send phases because they are being updated */
		if(SchedRun_Flag == TRUE)
			return 0;

		/* If first message, init arb table index and set new plan detection flag*/
		if( Master_SendInd == (TotMesgs_MsgProp+1) )
		{
			NewPlan_Flag = FALSE;
			cur_ind=0;
		}

		/* Store sched. info in buffer */
		ctrl_mesg_data[0] = arb_table.vars[cur_ind].init;
		ctrl_mesg_data[1] = arb_table.vars[cur_ind].vid;
		cur_ind++;
		nb_ctrl_mesg=2;
					
		if(cur_ind<arb_table.nvars)
		{	/* If there are more messages(there is place for 4 in each MST_DATA_SCHINF (RM) frame */
			ctrl_mesg_data[2] = arb_table.vars[cur_ind].init;
			ctrl_mesg_data[3] = arb_table.vars[cur_ind].vid;
			cur_ind++;
			nb_ctrl_mesg=4;
		}
			
		if(cur_ind<arb_table.nvars)
		{	/* If there are more messages(there is place for 4 in each MST_DATA_SCHINF (RM) frame */
			ctrl_mesg_data[4] = arb_table.vars[cur_ind].init;
			ctrl_mesg_data[5] = arb_table.vars[cur_ind].vid;
			cur_ind++;
			nb_ctrl_mesg=6;
		}
				
		if(cur_ind<arb_table.nvars)
		{	/* If there are more messages(there is place for 4 in each MST_DATA_SCHINF (RM) frame */
			ctrl_mesg_data[6] = arb_table.vars[cur_ind].init;
			ctrl_mesg_data[7] = arb_table.vars[cur_ind].vid;
			cur_ind++;
			nb_ctrl_mesg=8;
		}
		
		/* Txmit control message	*/
		CANADR = CAN_TX | AutoInc;
		/* 1st byte - ID10..ID3 */	
		CANDAT = (CTRL_MESG_ID_HP | MST_DATA_SCHINF) >> 3;
		
		/* 2nd byte : ID2..ID0 + RTR + DLC */
		CANDAT = ( MST_DATA_SCHINF << 5 ) | nb_ctrl_mesg;

		/* Queue data bytes */
		for(i=0;i<nb_ctrl_mesg;i++)		
			CANDAT = ctrl_mesg_data[i];

		/* Set TR bit (Transmission Request) */
		CANADR = CAN_CMR;
		CANDAT &= 0xF1;    /* keep high 4 bits, TR is read as 1, reset others */

		Master_SendInd++; /* Update index */

		return 0;

	} /* Txmit of MST_DATA_MSG_SCHINF message */


} /* Function send_mst_data */


/***************************************************************************
*
* get_network_status
*
* Returns the status of the network:
*		NETWORK_IDLE - no traffic is detected;
*		NETWORK_OPOK - normal FTT-CAN traffic detected (active Master present);
*
****************************************************************************/
unsigned char get_network_status(void)
{
	xdata unsigned int TMR_counter;

	/* Wait (at most) for the duration of one EC for the reception of a trigger message */
	TMR_counter = 0;
	FTT_EC_detected = FALSE;
	do
	{
		TMR_counter++;
		/**************************************************************************
			Note:	incrementing the TMR_counter variable requires (at least):
								MOVX 	A,@DPTR 	(2 cyc)	
								INC 	A 				(1 cyc)
								MOVX	@DPRT,A 	(2 cyc)
						
				    the exit condition requires (at least)
								MOVX 	A,@DPTR 	(2 cyc)
								ANL		A,direct	(1 cyc)									
								Jx							(2 cyc)
								MOV		A,Rr			(1 cyc)
								Jx							(2 cyc)
						
						Even with compil. optimisations, each run will require
							at least 5 machine cyc (increment) + 3 machine cyc (comp) = 8 cyc.
						With a clock of (aprox) 12MHz, results a minimum time of 8us.
						In each millisecond, the variable is incremented by 1000/8 = 125
		****************************************************************************/
	}while(FTT_EC_detected == FALSE && TMR_counter < (EC_E*125));

	/* Bus scan complete. Return result */
	if(FTT_EC_detected == FALSE)
		return NETWORK_IDLE;	/* No TM detected - network is idle */
	else if(FTT_EC_detected == TRUE)
		return NETWORK_OPOK;	/* TM detected - there is already an active master */
}


/**********************************************************
* S1_int
*
* Rotina de servio ao interrupt CAN
* (Serial port 1)
*
*   interrupt service routine
*
*
***********************************************************/

interrupt void S1_int(void)
{

	data unsigned char interrupts;

	/* Read CANCON register (read- int register) */
	interrupts = CANCON;

	/*******************************************/
	/* Do nothing if error or wakeup interrupt */
	/*******************************************/
	#ifdef DEVELOP
		if (interrupts & 0x04) {};     /* error - EI */    /* Disabled */
		if (interrupts & 0x10) {};     /* wake up - WUI */
	#endif

	/********************************************/
	/* Overrun error of reception buffer, do ?? */
	/********************************************/
	if ( interrupts & 0x08 )     /* overrun - OI */
	{
		/* CMR: clear overrun status COS */
		CANADR = CAN_CMR;
		CANDAT &= 0xf8;  /* COS is read as 1, high bits must be kept */
		/* follow down to receive last message, forget previous one */
		/* rx int. must be pending too */
	}

	/**************************************/
	/* CAN Rx INT : message just received */
	/**************************************/
	if( interrupts & 0x01 )     /* RI - rx int. */
	{
		/* Read the full frame of the master message into IntMesgBuf[]  */
		/* Use DMA (buffer is in data memory (internal)                 */
		/* Set internal RAM add in CANSTA                               */
		CANSTA = (unsigned char) IntMesgBuf;
		/* start DMA to receive full frame (start at CAN_RX) */
		CANADR = CAN_RX | DMA;

		/* Make time for DMA transfer (2 instr. cycle) */
		U8051NOP
		U8051NOP

		/* CMR: release RX buffer, RRB */
		CANADR = CAN_CMR;

		CANDAT &= 0xf4;  /* RRB is read as 1, high bits must be kept */

		/***************************/
		/* If Trigger Message, ... */
		/***************************/
		if( (IntMesgBuf[0] & 0xF0) == (MST_MESG_ID>>3) )
		{
			FTT_EC_detected = TRUE; /* Signals the reception of a Trigger Message */
			
			switch(Master_Status)
			{
			
				case ACTIVE_MASTER: 					/* Master is an Active Master */
					/* Trigger Message received by the active Master - should not happen */
					Master_Status = BACKUP_MASTER_OUTSYNC; /* Error!. Get fresh data and restart as a Backup Master */
					Pol_Status = BM_POL_DATA;
					Master_SendSRT_Data = FALSE; /* Clear flag */

					TR0=0;	/*Stop Dispatcher (timer 0) ints  */

					break;
				
				case BACKUP_MASTER_OUTSYNC: 	/* Master is an Backup Master and it is out of sync */
					if(Pol_Status == BM_POL_DATA )
					{

						/* Txmit control message - MST_DATA_QRY	*/
						CANADR = CAN_TX | AutoInc;
						/* 1st byte - ID10..ID3 */
						CANDAT = (CTRL_MESG_ID_HP | MST_DATA_QRY) >> 3;

						/* 2nd byte : ID2..ID0 + RTR + DLC */
						CANDAT = ( MST_DATA_QRY << 5 ) | 0x00;

						/* Set TR bit (Transmission Request) */
						CANADR = CAN_CMR;
						CANDAT &= 0xF1;    /* keep high 4 bits, TR is read as 1, reset others */

						Pol_Status = BM_WAIT_DATA;
					}
					else
					{
						if(IntMesgBuf[0] & 0x01)
							NewPlan_Flag = TRUE; /* Set new plan flag if apprpriate */

					}
					
					break;
					
				case BACKUP_MASTER_GETSYNC: 	/* Master is an Backup Master and is waiting for the first TM for timing synchronization */
					/* Set initial value and start Timer 0. Note that there is the tx delay + processing delay */
					TH0=TM_ETE_delay_TH0_ticks;
					TL0=0;
					TR0=1;
						
					/* Update TM seq. number and Current EC index */
					/*    Note: this should be the second EC in the plan. The first EC is used to trigger the synch process.  */
					/*          Since less than 7 ECs are required the index to the */
					/*          dispatcher table can be taken from the TM seq. number */
					/*          of the TM sent by the active master */
					CurrentEC = (IntMesgBuf[1] >> 5)+1;
					tm_seqnum = CurrentEC;
						
					/* Signal that Backup Master is in sync. */
					Master_Status = BACKUP_MASTER_SYNC;
					
					break;
				
				case BACKUP_MASTER_SYNC: 	/* Master is an Backup Master and is synchronized */
					/* Set initial value and start Timer 0. Note that there is the tx delay + processing delay */
					/*   It is necessary to resynch because of clock drifts */
					TR0=0;  /* Stop Timer 0 */
					//TF0=0;	/* Clear T0 int flag */
					TH0=TM_ETE_delay_TH0_ticks;
					TL0=0;

					//for(dbg_wait_cnt=0;dbg_wait_cnt<1;dbg_wait_cnt++);


					TR0=1;


					/* Note : counters and state variables are updated in the Timer0 ISR */
					/*        This int should occur after the Timer 0 int */
					
					break;

			}
		} /* IF Trigger Message received */

		/**********************************/
		/* If Synchronous Data Message... */
    /**********************************/
		if( (IntMesgBuf[0] & ~((DMESG_MASK|TXNEWDAT_MASK)>>3)) == (DATA_MESG_ID>>3) )
		{
			/* Error detection can be placed here .... */

		} /* If Sync. data message received */

		/**********************************/
		/* If Asynchronous Data Message.. */
		/**********************************/
		if( (IntMesgBuf[0] & ~((DMESG_MASK|TXNEWDAT_MASK)>>3)) == (AM_DATA_MESG_ID>>3) )
		{

		} /* Async. Data Message Received */


		/*************************************/
		/* If Asynchronous Control Message.. */
		/*************************************/
		if( ((IntMesgBuf[0] & ~((DMESG_MASK|TXNEWDAT_MASK)>>3)) == (CTRL_MESG_ID_LP >> 3) )  || ( (IntMesgBuf[0] & ~((DMESG_MASK|TXNEWDAT_MASK)>>3)) == (CTRL_MESG_ID_HP>>3) ) )
		{
			/* Check if Master is an Active Master */
			if(Master_Status == ACTIVE_MASTER)
			{					
			
				/* Check the kind of control message */
				if( ( IntMesgBuf[0] & ((DMESG_MASK|TXNEWDAT_MASK)>>3) << 3) | (IntMesgBuf[1] >> 5 ) == MST_DATA_QRY)
				{
					if(Master_SendSRT_Data == FALSE)
					{
						Master_SendSRT_Data = TRUE;
						Master_SendInd = 0;	/* Signal that it is a new query */
					}

				}
				
			}
			
			/* Check if Master is an backup Master and is waiting for data*/
			if(Master_Status == BACKUP_MASTER_OUTSYNC && Pol_Status == BM_WAIT_DATA)
			{
			
				cm_type = ((IntMesgBuf[0] & 0x0F)<<3) | (IntMesgBuf[1] >> 5); /* Get control message type */
			
				/* If beginning of data download */
				if(cm_type == MST_DATA_MSGPROP_BEG)
				{
					/* Store number of messages that should be received */
					MstDataQueue.nm_srt = IntMesgBuf[2];	
					MstDataQueue.nm_msgprop = IntMesgBuf[3];	
					MstDataQueue.nm_schinf = IntMesgBuf[4];	
					
				}
			
				/* If message properties or scheduling info message */
				if(cm_type == MST_DATA_MSGPROP || cm_type == MST_DATA_SCHINF)
				{
					nbm = IntMesgBuf[1] & 0x0F; /* Get number of data bytes in mesg */
				
					for(i=0;i<nbm;i++) /* Store mesg info */
						MstDataQueue.mesg_data[MstDataQueue.nmesgs][i]=IntMesgBuf[i+2];
					MstDataQueue.mesg_nb[MstDataQueue.nmesgs]=nbm;	
					
					MstDataQueue.nmesgs++;	/* Update queue index */
		
				}
				
				/* If message signals data txfer complete and OK */
				if(cm_type == MST_DATA_OK)
				{
					NewPlan_Flag = FALSE; /* To detect next new plan event */
					Pol_Status = BM_ROK_DATA;
				}
				
				/* If message signals SCHINF refresh data needed */
				if(cm_type == MST_DATA_SCHINF_REFRESH)
				{
					MstDataQueue.nmesgs = MstDataQueue.nm_msgprop;	/* Update queue index to receive updated SCHIF data*/

				}
				
			}
		
			
		} /* Async.Control Message Received */


	} /* RX Int */

	
	/**********************************/
	/* CAN Tx INT : message just sent */
	/**********************************/
	if( interrupts & 0x02 )     /* tx int. - TI */
	{
		/* TX ints are enabled. The txmition of the TM will generate a TX int. */
		if( Master_Status == ACTIVE_MASTER && Master_SendSRT_Data == TRUE) /* If data was requested */
		{
			if(TH0 < TH0_asylen)	/* If in async. window */
				send_mst_data();	/* Send next message (one each time) */

		}


		/**** DEBUG CODE - STOP AFTER A NEW BACKUP MASTER ENTERS ***/
		/* Force Master change */
		if( Master_Status == ACTIVE_MASTER && dbg_parar==TRUE)
		{
			if(dbg_parar_cnt > 100)
				{
					EA=0;
					while(1);
				}

		}
		/***** DEBUG CODE END **********/

		
		if( Master_Status == BACKUP_MASTER_SYNC) /* Backup Master syncronized with the active master */
		{
			/* The backup master txmits a TM with a small delay relatively to the active master, and then aborts it */
			/* If the active master is operating OK, the TM will be aborted (busy bus)   */
			/*		else the TM starts txmit immediatly and the abort command fails        */
			
			if( (CANSTA & 0x08) ) /* Stat_Reg:Bit3 - Tx Complet. Status (1-Last TX OK ;0-Last TX Fail) */
			{
				/* Abort command fail -> Last Tx request OK - > Become the active Master */			
				Master_Status = ACTIVE_MASTER;
				Master_SendSRT_Data = FALSE;
				
			}
			else
			{	/* Abort command sucessful */
			
			}
		
		}		
	
	} /* IF Tx INT */

  /* default return */
  return;
}



/*******************************************
*  T0_int
*  Timer 0 interrupt service routine
*
*    dispatcher
*
*  scans current plan, one EC per call
*  builds respective master message and
*  send it to the bus
*
********************************************/
interrupt void T0_int(void)
{
	data unsigned char i;
	data unsigned char EC_vars;
	data unsigned char index;
	data unsigned char new_plan;
	data unsigned char smlen_500us; /* Length of synch. messages to produce in # of 500us slots */


	switch( disp_plan->status )
	{
		case OUT_OF_DATE:
			/* starting up, do nothing until disp_plan status = UPDATED */
			return;

		case MISSED_DEADLINE:

			/* missed deadline detected by scheduler, terminate */
			return;

		case BUSY_SCHED:
			/* scheduler did not build plan in time, terminate */
			//printf("\n *** Fatal Error : Scheduler has not finished in time\n");
			while(1){};
			return;

		case UPDATED:
			disp_plan->status = BUSY_DISP;   /* update status */
			/* follow down to normal plan dispatching */
			break;

		case BUSY_DISP:
			/* normal plan dispatching */
			/* perform phony ops to make exec time exactly equal to case UPDATED */
			disp_plan->status = BUSY_DISP;
			break;   /* follow on with the dispatcher code */

		default:
			/* strange error !!!,  terminate */
			return;
	} /*Switch */


	/* Send master message with codified data */
	/* TM Format: ID field = Type(4b) + MST_ID(3b) + NewPlan_flag(1b)+Frame_Serial_number(3b) */
	/* First data byte conveys the # of 500us slots needed by synch. */
	/*  messages in current EC                                       */
	/*
	if(TBS == 0) /* CAN Tx buff in use -ERROR*/
	{
		//return;
	}
	*/

	CANADR = CAN_TX | AutoInc;
	smlen_500us = disp_plan->EC[CurrentEC].load500us;
	if(CurrentEC == 0)
		new_plan=0x01;
	else
		new_plan=0x00;

	/* 1st byte - ID10..ID3 */
	CANDAT = (MST_MESG_ID >> 3) | (MST_ID << 1) | new_plan;
		
	/* 2nd byte : ID2..ID0 + RTR + DLC */
	CANDAT = ( tm_seqnum  << 5 ) | (MAX_VAR / 8 + 1);

	/* First data byte - Sync. window length in slots of 500us */
	CANDAT =	smlen_500us;

	/* Remaining data bytes - Trig. for sync mesgs */	
	for( i=0; i < MAX_VAR/8; i++ )
		CANDAT = disp_plan->EC[CurrentEC].EC_list[i];

	/* Set TR bit (Transmission Request) */
	CANADR = CAN_CMR;
	CANDAT &= 0xF1;    /* keep high 4 bits, TR is read as 1, reset others */
		
	/* Check if backup Master */
	if(Master_Status == BACKUP_MASTER_SYNC)
	{	
		/* Similar to Master_Status == ACTIVE_MASTER, except that it aborts the Txmition and */
		/*   based on the result of the abort may change the Master Status */

		/* Abort Tx of requested txmission */
		CANADR = CAN_CMR;
    CANDAT &= 0xF2;     /* Keep high 4 bits. AT is read as 1, reset others */
   											/* Note : a tx int will be generated independently of the result of the abort command */
   											/*        Therefore only is S1_int/Tx_int the status can be checked and appropriate actions taken */

	}	

	/* Store the value that TH0 will have at the end of the allowed transmition area of the asynchronous window */
	TH0_asylen = 255-(smlen_500us* SLOT500US_TICKS+MAXAMVAR_TICKS);
	/* Note: async. mesgs will be txmitted by the TX int event, following the txmition of the TM */

	/* Update TM seq. number, Current EC index and check end of plan */
	tm_seqnum++;
	if(tm_seqnum > 0x07)
		tm_seqnum = 0;
	
	CurrentEC++;
	if( CurrentEC == NEC )
	{
		/* End of plan reached, reset CurrentEC, switch plans */
		CurrentEC = 0;
		tm_seqnum = 0;
		disp_plan->status = OUT_OF_DATE;
		temp_plan = sched_plan;
		sched_plan = disp_plan;
		disp_plan = temp_plan;  /* set disp_plan last because of ints */

		NewPlan_Flag = TRUE;
		SchedRun_Flag = TRUE;
	}

	/* Just to provide sync for osciloscope */
	if( (CurrentEC % 4) == 0 )
	{
		P4.7 = 1;
		P4.7 = 0;
	}

	if(dbg_parar == TRUE)
		dbg_parar_cnt++;

	return;

}

/**********************************
*
* ini_CAN
*
* Inicializao do interface CAN
*
***********************************/
void ini_CAN(void)
{

  /* Inicializao do interface CAN */
  /* Criado com 80592.xls */
  /* Baud Rate = 122.88 kHz(@11.0592MHz), Interface=250 */

  /* Force Reset Request, transmit and receive CAN interrupts only*/
  CANADR = CAN_CR;
	CANDAT = 0x67;  /* TM=0, S=RA=1, OIE=EIE=0, TIE=0, RIE=1, RR=1 */
  CANADR = CAN_CMR;
  CANDAT = 0xC0;   /* RX0A=RX1A=1, WUM=SLP=0 */
  
  CANADR = CAN_ACR | AutoInc;
  CANDAT = 0x00;	/* ACR - not relevant (See AMR) */
  CANDAT = 0xff;	/* AMR - receive all messages */
  CANDAT = 0x44; 	/* BTR0 */
  CANDAT = 0xA4;	/* BTR1 */
  CANDAT = 0x1A;	/* OCR */
  
  /* Terminate Reset Request and keep other bits as they were */
  CANADR = CAN_CR;
  CANDAT &= 0xFE; /* RR=0 */

	/* Interrupt Setup  */
	ES1=1;  /* Enable SIO1 (CAN) Interrupt          */
	PS1=1;  /* Priority Level of SIO1 (CAN) = high  */
}


/**************************************
*
*     ini_globals
*
*  sets up globals and timer to
*  generate periodic interrupt
*  to control the EC duration
*
**************************************/
void ini_globals(void)
{

	Master_SendSRT_Data = FALSE; /* No request for SRT data is pending */
	Master_Status = BACKUP_MASTER_OUTSYNC; /* Default boot state : backup master with no SRT data */
	
	arb_table.nvars=0;

	plan_1.status = OUT_OF_DATE;
	plan_2.status = OUT_OF_DATE;

	sched_plan=&plan_1;
	disp_plan=&plan_2;
  
	CurrentEC = 0;
	
	tm_seqnum = 0;	/* Sequence number of the Trigger Message */
	
	/* globals to collect timing data */
	SchedTIndex=0;
	NintECs=0;
	DispTIndex=0;

	timer_ini(0,	/* Timer 0 */
	    			0,	/* Timer */
						0,	/* Modo = 0; int @112.5Hz (8.9ms) */
								/* Modo = 1; int @14.06Hz (71ms) */
						0,	/* Controlo s pelo bit TR0 */
						1,	/* Gera ints */
						0,	/* Priority */
						0); /* Recarga; N/A */

		ini_CAN();
	RS232_ini(0xfd,0,0);

}

/*****************************************************
*
* msg_set_add
*
* Adds a predefined set of messages to the SRT
******************************************************/
unsigned char msg_set_add(void)

{
	MstVarType aux_var;
	
	aux_var.vid=0x01;
	aux_var.size=1;
	aux_var.period=1;
	aux_var.deadline=1;
	aux_var.init=0;
	MstTab_AddVar(&aux_var);

	aux_var.vid=0x02;
	aux_var.size=3;
	aux_var.period=1;
	aux_var.deadline=1;
	aux_var.init=0;
	MstTab_AddVar(&aux_var);

	aux_var.vid=0x03;
	aux_var.size=3;
	aux_var.period=2;
	aux_var.deadline=2;
	aux_var.init=0;
	MstTab_AddVar(&aux_var);

	aux_var.vid=0x04;
	aux_var.size=2;
	aux_var.period=3;
	aux_var.deadline=3;
	aux_var.init=0;
	MstTab_AddVar(&aux_var);

	aux_var.vid=0x05;
	aux_var.size=5;
	aux_var.period=4;
	aux_var.deadline=4;
	aux_var.init=0;
	MstTab_AddVar(&aux_var);

	aux_var.vid=0x06;
	aux_var.size=5;
	aux_var.period=4;
	aux_var.deadline=4;
	aux_var.init=0;
	MstTab_AddVar(&aux_var);

}


/****************************************************************
*
*    MstTab_AddVar()
*
*   used in stations and arbiter
*   appends a variable to the master variables' table
*   fills in auxiliar fields: size_b, byte_index and bit_mask
*
*   inputs:
*      - pointer to struct with properties of variable to add:

*         vid (6bit short identifier)
*         size (data size in bytes)
*         period,deadline,init (in number of ECs)
*      - global variables' table (MstVTabType arb_table)
*   outputs:
*      - error code: 0 - OK
*                    -1 - table full (v_table->nvars = MAX_VAR)
*
*******************************************************************/
signed char MstTab_AddVar(MstVarType *aux_var)
{
  unsigned char aux_bit,aux_byte;

  	/* check if table is full, if yes - terminate with error */
 	if(arb_table.nvars==MAX_VAR)
   		return(-1);

 	/* Complete the input structure, fields: b_size, byte_index, bit_mask */
 	/* Calculate size in bits */
 	/* Note that in Version 8 only Tx time is taken in account */
 	/* because stations are Full-CAN or each station only Tx a single message in each EC*/
	aux_var->b_size= bits_OVRHD_sb + 8 * aux_var->size;
	aux_var->b_size += (int)ceil((double)aux_var->b_size / (double)5);
	aux_var->b_size += bits_OVRHD_nsb;

	printf("\nAddVar: ID:%d size:%d b_size:%d",aux_var->vid,aux_var->size,aux_var->b_size);


	/* calculate respective byte and bit for codification in master message */
  	/* get lower part of variable's identifier (always 6 bits) */
  	aux_bit=(aux_var->vid)&0x3f;

  	/* determine in which byte is the respective bit */
  	aux_byte=0;
  	while(aux_bit>7)
    {
    	aux_byte++;
		  /* build bit mask for that byte */
        aux_bit-=8;
	}
    /* set bit_mask and byte_index in the var data structure */
	 aux_var->byte_index=aux_byte;
    aux_var->bit_mask=1;
    if(aux_bit>0)
    	(aux_var->bit_mask)<<=aux_bit;

#ifdef EDF
	/* Set-up var table for EDF scheduling */
    if(aux_var->init == 0)
    {
    	aux_var->init = aux_var->period; /* Next instance of message */
		aux_var->absdeadline = aux_var->deadline-1;   /* Deadline for current instance */
    }
    else
    {
        aux_var->absdeadline = -1; /* Signal that no instance exists */

    }
#endif


    /* add var at the end of the table */
    arb_table.vars[arb_table.nvars]= *aux_var;

   	/* adjust number of variables in table */
    arb_table.nvars++;

    return(0);
}

/******************************************************************************
*
*    planning scheduler --  p_scheduler()
*
*   arbitrator scheduler routine
*   scans the arbitrator variables' table
*   builds a plan as a list a ECs
*   each EC has a list of variables to be produced in it
*   variables in each EC are identified by (byte_index,bit_mask)
*      this ideintification is equivalent to vid but facilitates dispatching
*
*   inputs:
*      - pointer to plan table (plan)
*      - global arbitrator's variables' table (arb_table)
*   outputs:
*      - error code:
*         0 - OK
*         MISSED_DEADLINE - self explanatory!

*      - a status code is used within the plan struct to
*          synchronize with the dispatcher
*
*******************************************************************************/
unsigned char p_scheduler(PlanType *plan)
{
  data unsigned char i,j,aux_char;
  data unsigned char byte_index,bit_mask;
  data int edf_vid,edf_absdeadline;
  data int aux_int,ECload;

/*  data unsigned char variab_absdeadline, variab_b_size;*/
  data unsigned char ps_nvars;

  data signed char index_to_EC,right_EC;
  data MstVarType *variab;
  data EcType *EC_aux;


	/* start to build next plan, status = BUSY */
	plan->status=BUSY_SCHED;

	ps_nvars=arb_table.nvars;

	 /* clear the plan table, (EC)list=0, (EC)load=0 */
	 for(i=0;i<NEC;i++)

	 {
		  for(j=0;j<MAX_VAR/8;j++)
			plan->EC[i].EC_list[j]=0;
		  plan->EC[i].load=0;
	 }

#ifdef RM
	/* Go through the variables' table */

	for(i=0;i<ps_nvars;i++)
	{
		variab= &(arb_table.vars[i]);
		/* check whether variable happens within the plan */
		if (variab->init>=NEC)
		{   /* next instance of variable is out of current plan */
			/* update initial phase for next plan */
			variab->init=variab->init-NEC;
		}
		else
		{   /* next instance of variable occurs in this plan */
			/* go through all instances that occur in the plan */
			for (right_EC=variab->init;right_EC<NEC;right_EC+=variab->period)
			{
				/* if variable came from previous plan right_EC(j=0)<0 */
				/* start scanning from 0 */
				if(right_EC<0)
					index_to_EC=0;
				else
					index_to_EC=right_EC;
				/*build pointer to right EC */
  				EC_aux=&(plan->EC[index_to_EC]);
  				/* check all ECs for room starting from the right one until
				**  deadline or end of plan */
  				do
				{
  					/* use aux_int to hold load */
					aux_int=EC_aux->load+variab->b_size;
  					if(aux_int>EC_MAX_bits)
  					{   /* variable does not fit in this EC, keep scanning */
  						EC_aux=&(plan->EC[++index_to_EC]);
						/* check if deadline reached */
  						if(index_to_EC-right_EC>=variab->deadline)
  						{   /* missed deadline, terminate (all deadlines are hard) */
							plan->status=MISSED_DEADLINE;
  							break;
  						}
  						/* check if reached end of plan */
						if(index_to_EC>=NEC)
  						{   /* end of plan, leave placement for next plan
  							**  this will be signalled with init<0 */
							break;
  						}
  					}
  				} while(aux_int>EC_MAX_bits);

  				/* if deadline was missed or end of plan was reached
  				**  break the for loop */
  				if(plan->status==MISSED_DEADLINE) break;
  				if(index_to_EC>=NEC) break;
				/* at this point, an EC with enough room was found,
				**  place variable in it */
				/* EC_aux points to EC with room, index_to_EC */
  				/* upddate EC load, aux_int has updated load already */
  				EC_aux->load=aux_int;
				/* add variable bit to EC_list (byte_index & bit_mask) */
  				/* use aux_char to hold byte index */
				aux_char=variab->byte_index;
  				EC_aux->EC_list[aux_char] |= variab->bit_mask;
			} /* for right_EC ... */

  			if(plan->status==MISSED_DEADLINE)
  				return(MISSED_DEADLINE);  /* terminate with error */

  			/* else, update init and pass on to next variable */
  			variab->init=right_EC-NEC;
  			/* notice that for init<0 right_EC<NEC (for loop broken) */
		} /* Variab->init <= NEC */
    }/* For i = 1 .. NVARS */

	/* Dynamic scanning window */
 	for(i=0;i<NEC;i++)
	{
		plan->EC[i].load500us = plan->EC[i].load/NBITS_500US;
		if( (plan->EC[i].load500us * NBITS_500US ) < plan->EC[i].load )
			++plan->EC[i].load500us;

	}

  	/* plan sucessfully built, clear status and return OK */
		plan->status=UPDATED;
		SchedRun_Flag=FALSE;
#endif

#ifdef EDF

	for(index_to_EC=0;index_to_EC<NEC;index_to_EC++) /* For each EC in the plan */
	{

		ECload = 0; /* EC load */

		/***** Scan message table while messages to produce are found */
		do
		{

			/* Get message with earliest deadline which fits in EC (if any) */
			edf_vid=-1;
			edf_absdeadline=-1;
			for(i=0;i<ps_nvars;i++) /* For each message */
			{
				variab=&arb_table.vars[i];
				/*variab_absdeadline = arb_table.vars[i].absdeadline;
				variab_b_size = arb_table.vars[i].b_size;
				*/
				if( (variab->absdeadline != -1) && (ECload+variab->b_size <= EC_MAX_bits) )
				{  /* If message instance is ready and fits */

					if(edf_vid == -1)
					{ /* If first message, store id and deadline */
						edf_vid=i;
						edf_absdeadline=variab->absdeadline;
					}
					else
					{
						if( edf_absdeadline > variab->absdeadline )
						{ /* If message with shorter deadline, store id and deadline */
							edf_vid=i;
							edf_absdeadline=variab->absdeadline;
						}
					}
				}
			}

			/* If message found, update data structures */
			if(edf_vid != -1)
			{ /* Message to produce was found */

				/* Add message to ECload */
				ECload+=arb_table.vars[edf_vid].b_size;

				/* Signal that message was scheduled */
				arb_table.vars[edf_vid].absdeadline=-1;

				/* Store message in plan */
				/*   Add variable bit to EC_list (byte_index & bit_mask) */
				/*   Update EC load */
				byte_index=arb_table.vars[edf_vid].byte_index;
				bit_mask=arb_table.vars[edf_vid].bit_mask;
				plan->EC[index_to_EC].EC_list[byte_index] |= bit_mask;


			}

		}while(edf_vid > -1);
		plan->EC[index_to_EC].load=ECload;

		/***** Check if any deadline missed and update deadlines */
		for(i=0;i<ps_nvars;i++) /* For each message */
		{
			if(arb_table.vars[i].absdeadline==0)
			{
				plan->status=MISSED_DEADLINE;
				return(MISSED_DEADLINE);	/* All missed deadlines are hard */
			}
			else
				if(arb_table.vars[i].absdeadline > 0)
					arb_table.vars[i].absdeadline--; /* Update deadline*/
		}

		/***** Update releases */
		for(i=0;i<ps_nvars;i++) /* For each message */
		{
			if(--arb_table.vars[i].init==0)
			{
				arb_table.vars[i].init=arb_table.vars[i].period;

				arb_table.vars[i].absdeadline=arb_table.vars[i].deadline-1;
			}
		}

        /***** Update number of #500us slots */
        plan->EC[index_to_EC].load500us = plan->EC[index_to_EC].load/NBITS_500US;
		if( (plan->EC[index_to_EC].load500us * NBITS_500US ) < plan->EC[index_to_EC].load )
			++plan->EC[index_to_EC].load500us;

	}

	/* Plan sucessfully built, clear status and return OK */
	plan->status=UPDATED;
	SchedRun_Flag=FALSE;

#endif
 	
 	return(0);


}

/************************************************
* Function that seds a message (DBG type)				*
* Used for debug only.                          *
*                                               *
* Input parameter: dbg_info (global var)        *
*       to identify the action (should be       *
*       assigned to a diferent value in each    *
*       invocation point.                       *
*                                               *
************************************************/
void Send_DBG_Mesg()
{

		while(TBS == 0);

		CANADR = CAN_TX | AutoInc;

		/* 1st byte - ID10..ID3 */
		CANDAT = (MST_MESG_ID_DBG >> 3) | (MST_ID << 1);

		/* 2nd byte : ID2..ID0 + RTR + DLC */
		CANDAT = 1;

		/*  First data byte - Sync. window length in slots of 500us */
		CANDAT =	dbg_info;

		/* Set TR bit (Transmission Request) */
		CANADR = CAN_CMR;
		CANDAT &= 0xF1;    /* keep high 4 bits, TR is read as 1, reset others */
}


