/********************************
*
*    Planning Scheduler for CAN
*
*      Monitor station
*
*  Scans the traffic in the network
*
*	 Stores data contents and timing info in an array and prints it at the end.
*
*  Uses an interrupt service routine
*
*  History:
*    98/10/6 - created by Luis de Almeida
*              using CAN libs developed by Pedro Fonseca  
*    98/10/13 - new buffer to collect timing data
*              collects data until buffer full
*              then sends data via RS232 and terminates
*              data to be collected:
*              - timer 0 reset every time master message is received
*              - time instants of all subsequent data messages
*                (buffer limited to 8 data message per Ec)
*                (resolution +/- 35 us)
*    98/11/4 - reception directly from RX buffer
*    98/12/18 - use internal RAM for global variables
*               uses DMA to read messages (full frame)
*  2000/02/10 - Adapted by Paulo Pedreiras to use in Version 7
*  2000/10/30 - Adapted by Paulo Pedreiras for use with FTT-CAN Version 8
*  2002/02/10 - Adapted by Paulo Pedreiras for use in FTT-CAN Version 9
*									Timer0 in mode 1 (16 bit timer);
*                 Overflow counter for extended range;
*									New types of messages supported;
*
***********************************************************************/


#include "io592.h"
#include <80592.h>
#include <special.h>
#include "plansch9.h"


void ini_CAN(void);
interrupt void S1_int(void);


/****************/
/* Global vars: */
/****************/

#define TIMEDATASIZ 300

data unsigned char mstdata[12]; /* received message (full frame) + Rx Time info*/

data unsigned char wait_flag;

unsigned char TimeData[TIMEDATASIZ][14];
data unsigned int TimeDataIndex;
data unsigned char DataMesgIndex;

data unsigned int TH0_OVR=0; /* Overflow counter for timer 0 */

/************************
*
*   main
*
*  initiate globals and CAN controller.
*  processing within an infinite loop (no interrupt).
*    
*
****************************/


void main(void)
{
	unsigned char i,nsm_bytes,mesg_type,dlc_bytes;
	unsigned int aux_int;

	wait_flag = 1; /* Signal that should not perfor data aquisition */

		/*******************/
	 /* Initializations */
	/*******************/

	timer_ini(  0,  /* Timer 0 */
							0,  /* Timer */
							1,  /* Modo = 0; int @112.5Hz (8.9ms) */
									/* Modo = 1; int @14.06Hz (71ms) */
							0,  /* Controlo s pelo bit TR0 */
							1,  /* Gera ints */
							0,  /* Priority */
							0); /* Recarga; N/A */

	timer_arranque(0,1); /* Arranque */

	RS232_ini(0xfd,0,0);

	printf("\n Monitor 9 : Initializing data structures .... ");
	for( TimeDataIndex=0; TimeDataIndex < TIMEDATASIZ; TimeDataIndex++)
		for( i=0; i < 10; i++)
			TimeData[TimeDataIndex][i]=0;
	printf(" COMPLETED!\n");

	TimeDataIndex = 0;

	ini_CAN();

		/**************************************************/
	 /* Collect data - processing done in S1_int isr   */
	/**************************************************/
	printf("\n Collecting data ...");
	while( TimeDataIndex < (TIMEDATASIZ-1) )
	{
			// printf("\n TimeDataIndex=%d",TimeDataIndex);
	}

	printf("\n Data collected!");
	printf("\n Preparing for data display ...");

	/* disable ints EA=0 to avoid corruption of memory */
	EA = 0;

	/* send data to RS232 */
	for( TimeDataIndex=0; TimeDataIndex < TIMEDATASIZ; TimeDataIndex++)
	{
		mesg_type = (TimeData[TimeDataIndex][0] & 0xf0) >> 4;
		if(mesg_type == (MST_MESG_ID >> 7)) /* Trigger message */
		{

			nsm_bytes= (TimeData[TimeDataIndex][1] & 0x0F) - 1;
			printf("\n TM  [%d %d %d]  >",TimeData[TimeDataIndex][10],TimeData[TimeDataIndex][11],TimeData[TimeDataIndex][12]);
			printf(" MstID:%x NewP:%x SeqN:%x",(TimeData[TimeDataIndex][0]&0x0F)>>1,TimeData[TimeDataIndex][0]&0x01,TimeData[TimeDataIndex][1]>>5);
			printf(" LSW(slots)=%d",TimeData[TimeDataIndex][2]);
			printf(" Msgs: ");
			for(i=0;i<nsm_bytes;i++)
				printf(" %x",(unsigned char)TimeData[TimeDataIndex][3+i]);
		}

		if(mesg_type == (MST_MESG_ID_DBG >> 7)) /* Shadow Trigger message */
		{

			nsm_bytes= (TimeData[TimeDataIndex][1] & 0x0F) - 1;
			printf("\n   >DBG  [%d %d %d]  >",TimeData[TimeDataIndex][10],TimeData[TimeDataIndex][11],TimeData[TimeDataIndex][12]);
			printf(" MstID:%x ",(TimeData[TimeDataIndex][0]&0x0F)>>1);
			printf(" DBG Info=%d",TimeData[TimeDataIndex][2]);
		}

		if(mesg_type == (CTRL_MESG_ID_HP >> 7)) /* Control Message (HP) */
		{

			dlc_bytes= TimeData[TimeDataIndex][1] & 0x0F;
			printf("\n CM_HP [%d %d %d] >",TimeData[TimeDataIndex][10],TimeData[TimeDataIndex][11],TimeData[TimeDataIndex][12]);
			printf(" Type:%x",((TimeData[TimeDataIndex][0] & 0x0f) << 3) | (TimeData[TimeDataIndex][1] >> 5 ) );
			printf("\n        Data Field:");
			for(i=0;i<dlc_bytes;i++)
				printf(" %x ",TimeData[TimeDataIndex][2+i]);
		}

	}

	/* terminate */
	while(1);
 
}





/************************************
*
*   ini_CAN
*
* Inicializao do interface CAN
*  ( com interrupo )
* 
*************************/

void ini_CAN(void)
{

    /* Inicializao do interface CAN */
    /* Criado com 80592.xls */
    /* Baud Rate = 122.88 kHz(@11.0592MHz), Interface=250 */

    /* force Reset Request, accept CAN tx and rx ints, only */
    CANADR = CAN_CR;
    CANDAT = 0x63;      /* TM=0, S=RA=1, OIE=EIE=0, TIE=0,RIE=1, RR=1 */
    CANADR = CAN_CMR;
    CANDAT = 0xC0;      /* RX0A=RX1A=1, WUM=SLP=0 */
  
    CANADR = CAN_ACR | AutoInc;
    CANDAT = 00;        /* ACR - does not matter */
    CANDAT = 0xff;      /* AMR- no bits are relevant */
    CANDAT = 0x44;      /* BTR0 */
    CANDAT = 0xA4;      /* BTR1 */
    CANDAT = 0x1A;      /* OCR */
  
    /* terminate Reset Request and keep other bits as they were */
    CANADR = CAN_CR;
    CANDAT &= 0xFE; /* RR=0 */

    /* Falta ligar os interrupts EA e ES1 */
    /* Fim da inicializacao CAN */

		IEN0 |= 0xA0;
  
}

/*******************************************
*  T0_int
*  Timer 0 interrupt service routine
*
*
********************************************/
interrupt void T0_int(void)
{
	TH0_OVR++;
	return;
}



/*********************************
* S1_int
*
* Monitoring routine 
* (Serial port 1)
*
*   monitor station interrupt service routine
*
*   each EC:
*   interpretes master message data
*   count all variables' messages
*   check if number is in accordance to
*      what is specified in the master message
*
*
****************************/

interrupt void S1_int(void)
{
    unsigned char i,j,aux_byte;
    unsigned char interrupts, highID, mesgDLC;

    /* Leitura do registo CANCON (read- int register) */
    interrupts = CANCON;

    /********************/
    /* message received */
    /********************/
    if( interrupts & 0x01 )     /* RI - rx int. */
    {
        /* read the full frame of the master message into mstdata[] */
        /* use DMA (mstdata is in data memory (internal) */
        /* set internal RAM add in CANSTA */
        CANSTA = (unsigned char) mstdata;
        /* start DMA to real full frame (start at CAN_RX) */
        CANADR = CAN_RX | DMA;

        /* make time !! */
        U8051NOP
        U8051NOP

        /* CMR: release RX buffer, RRB */
        CANADR = CAN_CMR;
				CANDAT &= 0xf4;  /* RRB is read as 1, high bits must be kept */

				TR0 = 0; /* Stop Timer0 to avoid inconsistencies between counters */
				TimeData[TimeDataIndex+1][10] = TH0_OVR;
				TimeData[TimeDataIndex+1][11] = TH0;
				TimeData[TimeDataIndex+1][12] = TL0;
				TR0 = 1; /* Restart Timer0 */


				/* if Master message, ... */
				if( (mstdata[0] & 0xF0) == (MST_MESG_ID>>3) )
				{

					if(wait_flag == 1) return; /* Wait for right moment to start */

					/* increment general buffer pointer */
					TimeDataIndex++;

					/* Save TM contents */
					i=(mstdata[1]&0x0F)+2;
					for(j=0;j<i;j++)
						TimeData[TimeDataIndex][j] = mstdata[j];

				}

				/* if shadow master message, ... */
				if( (mstdata[0] & 0xF0) == (MST_MESG_ID_DBG>>3) )
				{

					/* increment general buffer pointer */
					TimeDataIndex++;

					/* Save TM contents */
					i=(mstdata[1]&0x0F)+2;
					for(j=0;j<i;j++)
						TimeData[TimeDataIndex][j] = mstdata[j];

				}


        /* If synchronous data message ...*/
				if( (mstdata[0] & ~((DMESG_MASK | TXNEWDAT_MASK) >> 3)) == (DATA_MESG_ID>>3) )
				{
				}

        /* If asynchronous data message ...*/
				if( (mstdata[0] & ~((DMESG_MASK | TXNEWDAT_MASK) >> 3)) == (AM_DATA_MESG_ID>>3) )
				{
				}

				/* If control message (HP)...*/
				if( (mstdata[0] & ~((DMESG_MASK | TXNEWDAT_MASK) >> 3))  == (CTRL_MESG_ID_HP>>3) )
				{

					/* increment general buffer pointer */
					TimeDataIndex++;

					/* Save message contents */
					i=(mstdata[1]&0x0F)+2;
					for(j=0;j<i;j++)
						TimeData[TimeDataIndex][j] = mstdata[j];

					/* Start data aquisition */
					wait_flag=0;
				}

		}

		/* default return */
		return;
}
