/******************************************************************************************
*
*  Planning scheduler for CAN
*
*         header file - Version 9
*
*
*  History:
*  created 98/9/28
*     by Luis de Almeida (LdA)
*
*  98/9/29 - byte_index and bit_mask added
*            to the variable description struct
*  98/9/30 - synchronization between scheduler and 
*            dispatcher corrected. states OUT_OF_DATE
*            UPDATED, BUSY_DISP, BUSY_SCHED, added.
*  98/10/1 - types updated to maximize use of unsigned chars
*  98/10/15 - new EC and plan types EcType, PlanType. ECs are already
*            codified as data in the master message. Now
*            dispatching is easier and more accurate.
*  98/10/15 - ECs load takes into account queueing delays.
*  98/10/20 - prevIndex added to EcType to save the index of the
*            last var added to an EC (used by the scheduler to
*            calculate the max delay that the production of a var
*            can suffer.
*  98/11/1 - added definition for DATA_MESG_ID
*  98/11/4 - added status byte in v_type (to specify if PRODUCER or CONSUMER)
*  98/11/6 - model to calculate load has been updated
*            transmissions of data messages start after SCAN_WINDOW after rx of MM
*  98/12/30 - variables' structure contains full buffer for data value (8 bytes)
*  99/1/9 - creation of version 5. Many differences:
*           - just the low 6 bits of data messages vid's are saved
*             in the table
*           - 7th bit of the identifier reflects the TX_NEW_DAT bit.
*           - the stations' tables are different than the master's,
*             stations do not need many of the fields
*           - this separation requires 2 different functions 
*             to add vars to the tables
*           - in the master table, there is now room to save the
*             max jitter, min jitter, average estimation and auxiliar
*             terms for the scheduler execution time estimation
*           - *_AddVar() functions moved to *.c source files
*           - produce() and consume() functions added to *.c source files
*  99/1/21  - b_size in messages' structure turned to unsigned char (just
*             possible with TTX_K1/K2 reduced by using DMA)
*  99/2/1   - data double-buffering to cope with assynchronous production/
*             /consumption. Double-buffer in struct StaVarType
*  99/2/17  - creation of version 6. 
*           - Messages' (or variables') description tables (either in
*             the stations or in the master) always have the maximum size
*             (the size to hold MAX_VAR messages).
*           - The VID is used as an index to the messages' table. This
*             is used in reception to address a message quickly.
*           - Messages' are produced by stations as soon as possible after
*             reception of the master message (no SCAN WINDOW).
*           - A fast internal buffer is used to hold the byte_index and
*             bit_mask of the messges produced by each station
*  99/2/25  - addition of defs concerning the master configuration message
*             produced by the operator's console station
*  99/3/2   - the master configuration message has a speacial mask in the
*             most significant bits of the identifier to facilitate
*             reception by the master without extra CAN rx interrupts.
*  99/3/15  - REFRESH and PROMPT changed to RX_NEW_DAT and TX_NEW_DAT respectively.
*  00/02/01 - Version 7 developed jointly by Luis Almeida & Paulo Pedreiras.
*               Scanning window size dynamicaly adjusted in each EC so the
*               synchronous message area varies accordingly to the number and
*               size of the synchronous messages produced in each EC.
*	            i)  MST_MESG_ID modified to support synchronous message length codified
*                   in 4 lower bits. (500 us resolution at 122.9 Kb/s);
*               ii) EcType changed to acomodate load_500us, i.e., the number of 500us 
*	                slots needed to hold the number of sync. mesgs in each EC.
*  00/10/23 - Version 8 developed jointly by L. Almeida and P. Pedreiras.
*              		i) Production of asynchronous messages under controlled mode 
*
*  02/01/29 - Version 9 developed by L. Almeida, P. Pedreiras and J. Ferreira
*							Error handling added:
*							- Master redundancy
*							- Master synchronisation
*							- Message error handling
*
****************************************************************************************/


/*************************************************************************
** Scheme used to identify variables to be produced in one EC 
**
** Messages have identifiers formed by a lower order part,
** lower than MAX_VAR, and a higher order part controlled
** by the arbitrator.
**                 xxxxx   xxxxxxx  = 11 bit identifier
**                   |        |
**                  high     lower<MAX_VAR
** the higher part is not transmitted in the master message
** the lower part is codified in a single bit in the data field
** the position of the bit corresponds to the value of that part
**   MAX_VAR=32 (4 bytes), message id (lower part)=13
**   master message data field - xxxxxxxx xxxxxxxx xxXxxxxx xxxxxxxx
**                                                   |
**                         1- produce, 0- don't    bit 13
***************************************************************************/

/* General masks */
#define DMESG_MASK		0x3f  		/* mask for the low 6 bits of identifier */

#define TXNEWDAT_MASK	0x40  		/* mask for the 7th bit (bit 6) of identifier (TX_NEW_DAT bit) */

#define MST_MESG_ID		0x080  		/* master message id, low byte - 000 1xxx ssss       	*/

#define MST_MESG_ID_DBG 0x100		/* Master Message (shadow;not to be used) - debug purposes */
																/* notice that lower 4 bits ( ssss ) of id (11 bits) 	*/
																/* convey the synchronous message  length        			*/
#define DATA_MESG_ID 	0x300 		/* basic id of data messages, low byte - 001 0Pxx xxxx */
																/*           x varies from message to message */

#define AM_DATA_MESG_ID 0x600 	/* basic id of asynchronous data messages, low byte - 001 0Pxx xxxx */
															/*           x varies from message to message */
/* Control message masks */
#define CTRL_MESG_ID_HP	0x400		/* Frame ID of a control message (high priority) */
#define CTRL_MESG_ID_LP	0x780		/* Frame ID of a control message (low priority) */

#define MST_DATA_QRY 						0x7F 	/* Control message-request active master SRT data */
#define MST_DATA_MSGPROP_BEG		0x7E 	/* BeginningEnd of MST_DATA_MSGPROP message stream */
#define MST_DATA_MSGPROP 				0x7D 	/* Control message contents is id,,DLC,period and deadline */
#define MST_DATA_SCHINF 				0x7C 	/* Scheduler-dependent info (RM- id and initial phase) */
#define MST_DATA_OK							0x72 	/* All data was sent OK */
#define MST_DATA_SCHINF_REFRESH	0x71 	/* New plan during txfer - New data of SCHINF type will be sent */
#define MST_DATA_ABORT					0x70 	/* Abort current synchronization operation */

#define MST_DATA_DBG						0x01	/* Debug message */

/* Maximum sizes */
#define MAX_VAR 				16    	/* Max n. of variables (integer mult of 8) */
#define AM_STAMAXVAR  	8      	/* Max. number of async messages that a station can consume */
#define AM_BUF_LEN    	8      	/* Length of the asynchronous message TX buffer */

#define STA_MAX_VAR 		8    		/* max n. of variables in a single station */
#define EC_MAX_VAR  		10    	/* max. n. of vars in 1 EC, EC_MAX_bits/bits_OVRHD */
#define NEC  						30			/* actual n. of ECs in the plan */

#define MAX_MESGS_SRT_DATA 30		/* Maximum number of messages required to txfer message and scheduling data */
																/*   from the active master to a backup master */

/* Transmission parameters */
#define TX_RATE  			123     /* (122.9) transmission rate in Kbps */
#define bits_OVRHD_sb	(1+11+1+2+4+15)	/* Mess. overhead subjected to stuff bits */
																			/* 	SOF, ID,RTR,Reserved,DLC,CRC */

#define bits_OVRHD_nsb 	(3+2+7)	/* Mess. overhead not subjected to stuff bits */
																/* 	Interm.,ACK,EOF */

#define IntOverh			12	/* interrupt overhead of producer (in bits @ tx rate) */
#define MMDecOverh		60  /* master mesg decoding overhead at each station (in bits @ tx rate) */

#define EC_E   				9   	/*  (8.9) EC duration in ms */
#define EC_E_bits  		1094 	/* (8.9*122.9) EC_E*TX_RATE */

#define EC_MASTER_bits  72		/* (bits_OVRHD+MAX_VAR+MAX_VAR/5) - bit size of master mesg */


#define EC_E_us 				8900			/* EC duration in us */
#define EC_MASTER_us 		(EC_MASTER_bits*8.13) /* Trig mesg duration in microseconds */
#define CANT2_INTOVR_us 200		/* Interrupt overhead of S0 (CAN) int plus T2 int*/

#define EC_MESGWIN_us (EC_E_us-EC_MASTER_us-CANT2_INTOVR_us) /* EC duration minus len trig. mesg & int overheads */

#define EC_MESGWIN_T2t (unsigned int)(EC_MESGWIN_us*0.922)	/* Trig mesg duration in T2 ticks */
																														/* T2 source = fosc/12 -> 1us=11059/12000 T2 ticks */
#define EC_MESGWIN_T2tH (unsigned char)(EC_MESGWIN_T2t/255)
#define EC_MESGWIN_T2tL (unsigned char)(EC_MESGWIN_T2t%255)

#define SCAN_WINDOW  			14	/* window after rx of MM to scan table in timer ticks */
#define SCAN_WINDOW_bits 	60	/* SCAN_WINDOW*EC_E/256*TX_RATE, same in bit time */

/************************************************************************************/
/* Notes :  Bit time = 1 / Tx rate = 1 / 123K = 8.137us                             */
/*          THx_Ticks : TLx_Max * Count period = 2^5 * ( 1 / (1/12 * 11.059MHz))    */
/*            For each THx_Tick TLx must overflow (in mode 0)                       */
/************************************************************************************/
#define OVRHEAD_QUEU	5		/* Number of TH0 ticks (34.75us each) to set-up first synch. mesg */
													/* Note that if station is consumer of previous message, the overhead */

													/* of consuming this message must be included here                    */
													/* Also,if the station is the last producing an sync mesg and first to	*/
													/* produce an async mesg, this value should allow that the station		*/
													/* ends scanning the SRT */
#define OVRHEAD_QUEU_bits (OVRHEAD_QUEU*35/8) /* Idem, in bits @ TX_RATE */

#define EC_MASTER_TICKS 	15   	/* EC Master bits(70) * Bit_Time(8.137us)/TH_Tick_Time(34.75us) */
#define EC_MAX_bits  			(EC_E*TX_RATE-EC_MASTER_bits - MMDecOverh - SCAN_WINDOW_bits)  /* max. no. of bits within an EC for vars */
#define PLAN_W  					(EC_E*NEC)  /* Plan duration in ms */
#define SLOT500US_TICKS 	15   	/* Number of TH0 ticks (34.75us each) needed by a 500us slot */
#define TIMINT_CAN_OVERH 	1   	/* Number of TH0 ticks (34.75us each) for Timer,CAN and Mast. mesg. decode overhead */
#define NBITS_500US     	61   	/* Number of bits each slot of 500us can convey */
#define	T0_IntOVRH				2			/* Overhead of 3 T0 interrupts (with some processing */

#define MAXAMVAR_TICKS		40	/* TH ticks for an asynchronous 8 byte mesg */
															/* Message duration: (8 data bytes) -> 130bit * Tbit = 1100us ; 1100us/TH0_ticks = 32 */
															/* Note that in worst case a prior RX int (async mesg ) can delay abort command so */
															/* this time should be included */
														
#define TMDIS_OVR_us 			200	/* Time required to process the TM at the active master node */
#define TDEC_OVR_us				150	/* Time required to decode the TM at the backup master node */

/* Total time (in timer ticks) to produce a message (tx+tqueue) = TTX_K1+TTX_K2*size */
/* using DMA, notice that TTX_K1_bits+8*TTX_K2_bits < 256 */
#define TTX_K1				27		/* constant time in total time to tx and queue a message (in ticks) */
#define TTX_K1_bits 	116   /* TTX_K1*EC_E/256*TX_RATE, same in bit time */
#define TTX_K2 				4  		/* proportional const. (to n. bytes) in total time to tx and queue a message (in ticks) */
#define TTX_K2_bits 	17    /* TTX_K2*EC_E/256*TX_RATE, same in bit time */

/* Master configuration message */
#define COD_MESG_SIZ 	8  		/* data size of the configuration message */
														/* index (master table), period, deadline, init (relative), size, vid */
#define OP_CONS_ID  (MAX_VAR-1)	/* operator's console data message vid, period=NEC */
																/* carries modifications to the Master Var Table */
#define MaskBits_MASK 	0xC0
#define Change_MASK   	0xC0
#define Append_MASK   	0x80
#define Delete_MASK   	0x40

/* Status of the plan tables */
#define UPDATED 					0  		/* plan table built and ready to be used) */
#define OUT_OF_DATE 			1  		/* plan table dispatched */
#define BUSY_SCHED 				2  		/* plan table being used by the scheduler */
#define BUSY_DISP 				3  		/* plan table being used by the dispatcher */
#define MISSED_DEADLINE 	4  		/* hard deadline missed while scheduling */

/* Message status (do not use bit 7, see Consume(), Produce() or VarStatus() ) */
#define PRODUCER    	0x1 	/* current station produces this message */
#define CONSUMER    	0x2 	/* current station consumes this message */
#define TX_NEW_DAT  	0x4 	/* =1, value buffer has been refreshed (producer) */
#define RX_NEW_DAT  	0x8 	/* =1, value buffer has been refreshed (consumer) */
	/*  Next two use the same bit because are exclusive */
#define LASTBUF				0x10	/* (double buffer mode) indicates last buffer written 0/1 */
#define CORRUPTED			0x10 	/* (single buffer mode) signals possible corruption by rx/tx int */

/* Current window in the EC - SLAVES*/
#define ASY_WIN		0x00    /* Current phase of the EC is asynchronous window */
#define SYN_WIN		0x10    /* Current phase of the EC is synchronous window 	*/

/* AM_getbuf parameters */
#define GETDATA			0x00    /* Only get next buffer row  (if any) */
#define UPDATEINDEX	0x01    /* Update index, corresp. to get one buffer row */

/* AM_Produce & AM_getbuf return codes */
#define AMOK			 0	/* OK */
#define AMEMPTY		-1	/* AM_getbuf fail because buffer was empty */
#define AMBF			-2	/* AM_produce fail because buffer was full  */
#define AMIID			-3	/* AM_produce fail because Invalid Msg ID   */
#define AMIML			-4	/* AM_produce fail because Invalid Msg length   */

/* General constants */
#define TRUE		1
#define FALSE		0

/* Master Status */
#define ACTIVE_MASTER 				0	/* Master is active, i.e., it is the one that is transmit. the TM */
#define BACKUP_MASTER_OUTSYNC 1	/* Master is a backup master, but it out of sync */	
#define BACKUP_MASTER_SYNC		2	/* Master is a backup master in sync (can become an ACTIVE MASTER) */
#define BACKUP_MASTER_GETSYNC	3	/* Master is a backup master waiting for sync. */
	

/* Network Status */
#define NETWORK_IDLE 	0	/* No network traffic detected */
#define NETWORK_OPOK	1	/* Newtork is operating normally */

/* Control message types */
#define MST_DATA_QRY_TYPE 				0 	/* Control message-request of active Master SRT data */
#define MST_DATA_MSGPROP_TYPE 		1 	/* Control message contents is id,,DLC,period and deadline */
#define MST_DATA_MSGPROP_BEG_TYPE 2 	/* Beginning of MST_DATA_MSGPROP message stream */
#define MST_DATA_SCHINF_TYPE 			3 	/* Scheduler-dependent info (RM- id and initial phase) */
#define MST_DATA_ABORT_TYPE 			4 	/* Abort current synchronization operation */

/* Backup Master Pol Status */
#define BM_POL_DATA 	1	/* Will ask active master data */
#define BM_WAIT_DATA 	2	/* Is waiting for active master data */
#define BM_ROK_DATA		3	/* Master data received */

/* Inline definition of 8051 Nop instruction (for consume 1 instr. cycle after DMA) */
/* Note that the header file special.h must be included for _opc() use */
#define U8051NOP         _opc(0x00);

/*******************************************************************************
    Remember that TX_NEW_DAT bit is set (=1) when the value buffer is written
 by the application and is reset (=0) when the message is produced by
 the communication interface (producer station).
    The RX_NEW_DAT bit is set (=1) when the message is consumed by the
 communication interface and is reset (=0) when the respective
 value buffer is read by the application (consumer station).
    If the producing application fails to supply new data at the
 appropriate rate (message period) then the TX_NEW_DAT bit will be reset
 meaning that the same value is being transmitted again.
    If the consuming application reads from the interface faster than the
 message period, then the RX_NEW_DAT bit will remain reset for, at least,
 two consecutive reads meaning that the value has not been updated.
********************************************************************************/


/*******************************/
/* Master Data Structures  */
/*******************************/
/* Var-table row type*/
typedef struct
{
	unsigned char vid;  			/* variable's id (lower 6 bits) */
	unsigned char byte_index;	/* byte of master mess. data where this var. is codified */
	unsigned char	bit_mask;		/* bit within [byte_index] corresponding to var. */
	unsigned char	size;				/* n. of data bytes */
	unsigned char	b_size;			/* size in bits, b_size= TTX_K1_bits+TTX_K2_bits*size */
	unsigned char	period;			/* in n. of ECs */
	unsigned char	deadline;		/* message deadline, normally=period */
	signed char		init;				/* EC of first ocurrence */
	signed char		absdeadline;/* Used in EDF sched:number of EC's to deadline */
    												/*		If < 0 then no request is pending */

	unsigned char status;			/* PRODUCER, CONSUMER, TX_NEW_DAT, RX_NEW_DAT */
	unsigned char max_jit;		/* save max value of jitter (in # Ec's) */
	unsigned char	nsec;				/* worst-case response in ECs */
	unsigned char	occurmax;		/* max number of occurrences of var within a plan */
} MstVarType;

/* Master variables table type */
typedef struct
{
	unsigned char	nvars;
	MstVarType		vars[MAX_VAR];
} MstVTabType;


/* Queue of received data of a backup master */
typedef struct
{
	unsigned char nmesgs;  			/* Number of messages received */
	
	unsigned char nm_srt;				/* Number of messages in the active Master SRT */
	unsigned char nm_msgprop;		/* Number of MSGPROP messages required */
	unsigned char nm_schinf;		/* Number of SCHINF messages required */

	unsigned char mesg_data[MAX_MESGS_SRT_DATA][8];	/* Contents of messages */
	unsigned char mesg_nb[MAX_MESGS_SRT_DATA];			/* Number of data bytes of each message */
}MstDataQueueType;

/********************************/
/* Station Data Structures  */
/********************************/

/* Var-table row type */
typedef struct
{
	unsigned char vid;				/* variable's id (lower 6 bits) */
	unsigned char byte_index;	/* byte of master mess. data where this var. is codified */
	unsigned char bit_mask;		/* bit within [byte_index] corresponding to var. */
	unsigned char data0[8];		/* double-buffer to hold variable's value  */
	unsigned char data1[8];
	unsigned char size;				/* n. of data bytes */
	unsigned char status;			/* PRODUCER, CONSUMER, TX_NEW_DAT, RX_NEW_DAT, LASTBUF */
} StaVarType;

/* Station variables table type */
typedef struct
{
	unsigned char nvars;
	StaVarType 		vars[MAX_VAR];
} StaVTabType;

/* ***** AM messages to consume type */
typedef struct
{
	unsigned char vid;			/* variable's id (lower 6 bits) */
	unsigned char data0[8];	/* double-buffer to hold variable's value  */
	unsigned char data1[8];
	unsigned char size;			/* n. of data bytes */
	unsigned char status;		/* PRODUCER, CONSUMER, TX_NEW_DAT, RX_NEW_DAT, LASTBUF */
} AM_ConsStaVarType;

/* Station variables table type */
typedef struct
{
	unsigned char 		nvars;
	AM_ConsStaVarType vars[AM_STAMAXVAR];
} AM_ConsStaVTabType;


/* ***** Asynchronous message buffer var type */
typedef struct
{
	unsigned char AM_data[10];	/* Data to be transmitted: Msg_Id+rtr+Dlc+data*/
} AM_BufStaVarType;

typedef struct
{
	unsigned char 		AM_putind;        /* Index of insertion point in table  */
	unsigned char 		AM_getind;        /* Index of removing point in table */
	AM_BufStaVarType 	AM_vars[AM_BUF_LEN];
} AM_BufStaVTabType;

    /*********************************************************************************
		*	The asynchronous data table will be managed as a circular buffer.
    *    New messages will be added in position AM_putind and removed from AM_getind.
    *    If AM_putind == AM_getind the buffer is empty.
    *    When any index crosses AM_BUF_LEN it restarts in zero.
    **********************************************************************************/


/***********************/
/* ECs' type           */
/***********************/
typedef struct
{
	int load;									/* present EC load in tx bits */
	unsigned char load500us;	/* present EC load in #500us slots */
	unsigned char EC_list[MAX_VAR/8];
} EcType;
	/**********************************************************
	* Note: EC_list already contains the bit codified
	*       identification of the variables to be produced
	*       ready to send as master message data
	***********************************************************/

/**************************/
/* Plan table type        */
/**************************/
typedef struct
{
	unsigned char status;
	EcType EC[NEC];
} PlanType;


